import asyncio
import logging
import random
from contextlib import asynccontextmanager
from typing import AsyncIterator

import aiohttp
from aiohttp import ClientConnectorError, web

from ai.backend.logging import BraceStyleAdapter
from ai.backend.wsproxy.exceptions import ContainerConnectionRefused, WorkerNotAvailable
from ai.backend.wsproxy.types import RouteInfo

from .abc import AbstractBackend, HttpRequest

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

CHUNK_SIZE = 1 * 1024 * 1024  # 1 KiB


class HTTPBackend(AbstractBackend):
    routes: list[RouteInfo]

    def __init__(self, routes: list[RouteInfo], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.routes = routes

    @property
    def selected_route(self) -> RouteInfo:
        if len(self.routes) == 0:
            raise WorkerNotAvailable
        elif len(self.routes) == 1:
            selected_route = self.routes[0]
            if selected_route.traffic_ratio == 0:
                raise WorkerNotAvailable
        else:
            ratios: list[float] = [r.traffic_ratio for r in self.routes]
            selected_route = random.choices(self.routes, weights=ratios, k=1)[0]
        return selected_route

    def get_x_forwarded_proto(self, request: web.Request) -> str:
        return request.headers.get("x-forwarded-proto") or "http"

    def get_x_forwarded_host(self, request: web.Request) -> str | None:
        return request.headers.get("x-forwarded-host") or request.headers.get("host")

    @asynccontextmanager
    async def request_http(
        self, route: RouteInfo, request: HttpRequest
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        base_url = f"http://{route.kernel_host}:{route.kernel_port}"
        headers = dict(request.headers)

        if headers.get("Transfer-Encoding", "").lower() == "chunked":
            del headers["Transfer-Encoding"]
        async with aiohttp.ClientSession(
            base_url=base_url,
            auto_decompress=False,
        ) as session:
            async with session.request(
                request.method,
                request.path,
                headers=headers,
                data=request.body,
            ) as response:
                yield response

    @asynccontextmanager
    async def connect_websocket(
        self, route: RouteInfo, request: web.Request, protocols: list[str] = []
    ) -> AsyncIterator[aiohttp.ClientWebSocketResponse]:
        base_url = f"http://{route.kernel_host}:{route.kernel_port}"
        async with aiohttp.ClientSession(
            base_url=base_url,
            auto_decompress=False,
        ) as session:
            log.debug("connecting to {}{}", base_url, request.rel_url)
            async with session.ws_connect(request.rel_url, protocols=protocols) as ws:
                log.debug("connected")
                yield ws

    async def proxy_http(self, request: web.Request) -> web.StreamResponse:
        protocol = self.get_x_forwarded_proto(request)
        host = self.get_x_forwarded_host(request)
        remote_host, remote_port = (
            request.transport.get_extra_info("peername") if request.transport else None,
            None,
        )
        headers = {
            **request.headers,
            "x-forwarded-proto": protocol,
        }
        if self.circuit.app == "rstudio":
            headers["x-rstudio-proto"] = protocol
        if host:
            headers["forwarded"] = f"host={host};proto={protocol}"
            headers["x-forwarded-host"] = host
            if self.circuit.app == "rstudio":
                headers["x-rstudio-request"] = f"{protocol}://{host}{request.path or ''}"
            split = host.split(":")
            if len(split) >= 2:
                headers["x-forwarded-port"] = split[1]
            elif remote_port:
                headers["x-forwarded-port"] = remote_port
        if remote_host:
            headers["x-forwarded-for"] = f"{remote_host[0]}:{remote_host[1]}"
        upstream_request = HttpRequest(
            request.method,
            request.rel_url,
            headers,
            request.content,
        )
        route = self.selected_route
        log.debug(
            "Proxying {} {} HTTP Request to {}:{}",
            request.method,
            request.rel_url,
            route.kernel_host,
            route.kernel_port,
        )

        try:
            async with self.request_http(route, upstream_request) as backend_response:
                response = web.StreamResponse(
                    status=backend_response.status,
                    headers={**backend_response.headers, "Access-Control-Allow-Origin": "*"},
                )
                await response.prepare(request)
                async for data in backend_response.content.iter_chunked(CHUNK_SIZE):
                    await response.write(data)
                await response.drain()

                return response
        except aiohttp.ClientOSError as e:
            raise ContainerConnectionRefused from e
        except:
            log.exception("")
            raise

    async def proxy_ws(self, request: web.Request) -> web.WebSocketResponse:
        stop_event = asyncio.Event()

        async def _proxy_task(
            left: web.WebSocketResponse | aiohttp.ClientWebSocketResponse,
            right: web.WebSocketResponse | aiohttp.ClientWebSocketResponse,
            tag="(unknown)",
        ) -> None:
            try:
                async for msg in left:
                    match msg.type:
                        case aiohttp.WSMsgType.TEXT:
                            await right.send_str(msg.data)
                        case aiohttp.WSMsgType.BINARY:
                            await right.send_bytes(msg.data)
                        case aiohttp.WSMsgType.PING:
                            await right.ping(msg.data)
                        case aiohttp.WSMsgType.PONG:
                            await right.pong(msg.data)
                        case aiohttp.WSMsgType.CLOSE:
                            log.debug("{}: websocket closed", tag)
                            await right.close(code=msg.data)
                        case aiohttp.WSMsgType.ERROR:
                            log.debug("{}: websocket closed with error", tag)
                            await right.close()
                        case _:
                            log.debug("{}: Unhandled message type {}", tag, msg.type)
            except ConnectionResetError:
                pass
            except Exception:
                log.exception("")
                raise
            finally:
                log.debug("setting stop event")
                stop_event.set()

        route = self.selected_route
        log.debug(
            "Proxying {} {} WS Request to {}:{}",
            request.method,
            request.path or "/",
            route.kernel_host,
            route.kernel_port,
        )

        if "Sec-WebSocket-Protocol" in request.headers:
            protocols = list(request.headers["Sec-WebSocket-Protocol"].split(","))
        else:
            protocols = []

        downstream_ws = web.WebSocketResponse(protocols=protocols)
        await downstream_ws.prepare(request)

        try:
            async with self.connect_websocket(route, request, protocols=protocols) as upstream_ws:
                try:
                    async with asyncio.TaskGroup() as group:
                        group.create_task(
                            _proxy_task(upstream_ws, downstream_ws, tag="(up -> down)")
                        )
                        group.create_task(
                            _proxy_task(downstream_ws, upstream_ws, tag="(down -> up)")
                        )
                        log.debug("created tasks, now waiting until one of two tasks end")
                        await stop_event.wait()
                finally:
                    log.debug("tasks ended")
                    if not downstream_ws.closed:
                        await downstream_ws.close()
                    if not upstream_ws.closed:
                        await upstream_ws.close()
            log.debug("websocket connection closed")
        except ClientConnectorError:
            log.debug("upstream connection closed")
            if not downstream_ws.closed:
                await downstream_ws.close()
        return downstream_ws



import asyncio
import functools
import logging
import socket

from aiohttp import web

from ai.backend.logging import BraceStyleAdapter
from ai.backend.wsproxy.defs import RootContext
from ai.backend.wsproxy.proxy.backend import TCPBackend
from ai.backend.wsproxy.types import (
    Circuit,
    RouteInfo,
)

from .abc import AbstractFrontend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class TCPFrontend(AbstractFrontend[TCPBackend, int]):
    servers: list[asyncio.Server]
    server_tasks: list[asyncio.Task]

    root_context: RootContext

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.servers = []
        self.server_tasks = []

    async def start(self) -> None:
        config = self.root_context.local_config.wsproxy
        port_start, port_end = config.bind_proxy_port_range
        for port in range(port_start, port_end + 1):
            service_host = config.bind_host
            sock = socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            # sock.bind() can be a blocking call only if
            # we're trying to bind to a UNIX domain file or host is not an IP address
            # so we don't have to wrap bind() call by run_in_executor()
            sock.bind((service_host, port))
            server = await asyncio.start_server(
                functools.partial(self.pipe, port),
                sock=sock,
            )
            self.servers.append(server)
            self.server_tasks.append(asyncio.create_task(self._listen_task(port, server)))
        log.info(
            "accepting proxy requests from {}:{}~{}",
            config.bind_host,
            port_start,
            port_end,
        )

    async def _listen_task(self, circuit_key: int, server: asyncio.Server) -> None:
        try:
            async with server:
                await server.serve_forever()
        except Exception:
            log.exception("TCPFrontend._listen_task(c: {}): exception:", circuit_key)
            raise

    async def stop(self) -> None:
        for task in self.server_tasks:
            task.cancel()
            await task
        for server in self.servers:
            server.close()
            await server.wait_closed()

    def ensure_credential(self, request: web.Request, circuit: Circuit) -> None:
        # TCP does not support authentication
        return

    async def initialize_backend(self, circuit: Circuit, routes: list[RouteInfo]) -> TCPBackend:
        return TCPBackend(routes, self.root_context, circuit)

    async def update_backend(self, backend: TCPBackend, routes: list[RouteInfo]) -> TCPBackend:
        backend.routes = routes
        return backend

    async def terminate_backend(self, backend: TCPBackend) -> None:
        return

    async def pipe(
        self, circuit_key: int, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        backend: TCPBackend | None = self.backends.get(circuit_key)
        if not backend:
            writer.close()
            await writer.wait_closed()
            return

        try:
            await backend.bind(reader, writer)
        except Exception:
            log.exception("TCPFrontend.pipe(k: {}):", circuit_key)
            raise

    def get_circuit_key(self, circuit: Circuit) -> int:
        return circuit.port



import asyncio
import importlib
import json
import logging
from pathlib import Path
from typing import Any

import aiohttp_cors
import click
import tomlkit
from aiohttp import web
from setproctitle import setproctitle

from ai.backend.logging import LogLevel

from ..config import ServerConfig, generate_example_json
from ..openapi import generate_openapi
from ..utils import ensure_json_serializable
from .context import CLIContext

log = logging.getLogger(__spec__.name)  # type: ignore[name-defined,union-attr]


@click.group(invoke_without_command=False, context_settings={"help_option_names": ["-h", "--help"]})
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=click.Path(
        file_okay=True,
        dir_okay=False,
        exists=True,
        path_type=Path,
    ),
    default=None,
    help="The config file path. (default: ./manager.conf and /etc/backend.ai/manager.conf)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Set the logging level to DEBUG",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    config_path: Path,
    log_level: LogLevel,
    debug: bool,
) -> None:
    """
    Backend.AI WSProxy CLI
    """
    setproctitle("backend.ai: wsproxy.cli")
    if debug:
        log_level = LogLevel.DEBUG
    ctx.obj = ctx.with_resource(CLIContext(config_path, log_level))


@main.command()
@click.option(
    "--output",
    "-o",
    default="-",
    type=click.Path(dir_okay=False, writable=True),
    help="Output file path (default: stdout)",
)
def generate_example_configuration(output: Path) -> None:
    """
    Generates example TOML configuration file for Backend.AI Proxy Coordinator.
    """
    generated_example = generate_example_json(ServerConfig)
    if output == "-" or output is None:
        print(tomlkit.dumps(ensure_json_serializable(generated_example)))
    else:
        with open(output, mode="w") as fw:
            fw.write(tomlkit.dumps(ensure_json_serializable(generated_example)))


async def _generate() -> dict[str, Any]:
    from ..server import global_subapp_pkgs

    cors_options = {
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=False, expose_headers="*", allow_headers="*"
        ),
    }

    subapps: list[web.Application] = []
    for subapp in global_subapp_pkgs:
        pkg = importlib.import_module("ai.backend.wsproxy.api" + subapp)
        app, _ = pkg.create_app(cors_options)
        subapps.append(app)
    return generate_openapi("Proxy Coordinator", subapps, verbose=True)


@main.command()
@click.option(
    "--output",
    "-o",
    default="-",
    type=click.Path(dir_okay=False, writable=True),
    help="Output file path (default: stdout)",
)
def generate_openapi_spec(output: Path) -> None:
    """
    Generates OpenAPI specification of Backend.AI API.
    """
    openapi = asyncio.run(_generate())
    if output == "-" or output is None:
        print(json.dumps(openapi, ensure_ascii=False, indent=2))
    else:
        with open(output, mode="w") as fw:
            fw.write(json.dumps(openapi, ensure_ascii=False, indent=2))


if __name__ == "__main__":
    main()
